﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;

namespace Template {

public struct TrainingEntry
{
	public float[] inputs;
	public int[] expected;
};
public struct TrainingSet
{
	public TrainingEntry[] entry;
	public int size;
};
public struct TrainingData
{
	public TrainingData( int t, int g, int v )
	{
		trainingSet.entry = new TrainingEntry[t];
		trainingSet.size = t;
		generalizationSet.entry = new TrainingEntry[g];
		generalizationSet.size = g;
		validationSet.entry = new TrainingEntry[v];
		validationSet.size = v;
	}
	public TrainingSet trainingSet;
	public TrainingSet generalizationSet;
	public TrainingSet validationSet;
};
public class NeuralNet
{
	// constants defining the topology of the net
	const int INPUTSIZE = 28 * 28;
	const int NUMHIDDEN = 150;
	const int NUMOUTPUT = 10;
	// constants steering training process
	const float LEARNINGRATE = 0.001f;
	const float MOMENTUM = 0.9f;
	const int MAXEPOCHS = 600;
	const float TARGETACCURACY = 95.0f;
	// pi
	const float PI = 3.14159265359f;
	// neural net storage
	float[] inputNeurons;
	float[] hiddenNeurons;
	float[] outputNeurons;
	int[] clampedOutputs;
	float[] weightsInputHidden;
	float[] weightsHiddenOutput;
	// training data
	TrainingData trainingData;
	float[] deltaInputHidden;		
	float[] deltaHiddenOutput;		
	float[] errorGradientsHidden;	
	float[] errorGradientsOutput;	
	int currentEpoch;			
	float trainingSetAccuracy;
	float validationSetAccuracy;
	float generalizationSetAccuracy;
	float trainingSetMSE;
	float validationSetMSE;
	float generalizationSetMSE;
	// rng
	Random rd = new Random( 1000 );
	// timer
	Stopwatch timer = new Stopwatch();
	public NeuralNet()
	{
		InitializeNetwork();
		InitializeWeights();
	}
	int GetInputHiddenWeightIndex( int inputIdx, int hiddenIdx ) { return inputIdx * NUMHIDDEN + hiddenIdx; }
	int GetHiddenOutputWeightIndex( int hiddenIdx, int outputIdx ) { return hiddenIdx * NUMOUTPUT + outputIdx; }
	float SigmoidActivationFunction( float x ) { return 1.0f / (1.0f + (float)Math.Exp( -x )); }
	int ClampOutputValue( float x ) { if ( x < 0.1f ) return 0; else if ( x > 0.9f ) return 1; else return -1; }
	float GetOutputErrorGradient( float desiredValue, float outputValue ) { return outputValue * (1.0f - outputValue) * (desiredValue - outputValue); }
	void InitializeNetwork()
	{	
		const int totalNumInputs = INPUTSIZE + 1;
		const int totalNumHiddens = NUMHIDDEN + 1;
		inputNeurons = new float[totalNumInputs];
		hiddenNeurons = new float[totalNumHiddens];
		outputNeurons = new float[NUMOUTPUT];
		clampedOutputs = new int[NUMOUTPUT];
		// set bias values
		inputNeurons[INPUTSIZE] = -1.0f;
		hiddenNeurons[NUMHIDDEN] = -1.0f;
		// allocate storage and initialize and layer weights
		weightsInputHidden = new float[totalNumInputs * totalNumHiddens];
		weightsHiddenOutput = new float[totalNumHiddens * NUMOUTPUT];
		// allocate storage for training
		deltaInputHidden = new float[(INPUTSIZE + 1) * (NUMHIDDEN + 1)];
		deltaHiddenOutput = new float[(NUMHIDDEN + 1) * NUMOUTPUT];
		errorGradientsHidden = new float[NUMHIDDEN + 1];
		errorGradientsOutput = new float[NUMOUTPUT];
	}
	float NormalDistribution( double stddev )
	{
		double u1 = 1.0 - rd.NextDouble(), u2 = 1.0 - rd.NextDouble();
		double randStdNormal = Math.Sqrt( -2.0 * Math.Log( u1 ) ) * Math.Sin( 2.0 * Math.PI * u2 );
		return (float)(stddev * randStdNormal);
	}
	void InitializeWeights()
	{
		const float distributionRangeHalfWidth = 2.4f / INPUTSIZE;
		const float standardDeviation = distributionRangeHalfWidth * 2 / 6;
		// set weights to normally distributed random values between [-2.4 / numInputs, 2.4 / numInputs]
		for( int i = 0; i <= INPUTSIZE; i++ ) for( int j = 0; j < NUMHIDDEN; j++ )
		{
			int weightIdx = GetInputHiddenWeightIndex( i, j );
			float weight = NormalDistribution( standardDeviation );
			weightsInputHidden[weightIdx] = weight;
		}
		// set weights to normally distributed random values between [-2.4 / numInputs, 2.4 / numInputs]
		for( int i = 0; i <= NUMHIDDEN; i++ ) for( int j = 0; j < NUMOUTPUT; j++ )
		{
			int weightIdx = GetHiddenOutputWeightIndex( i, j );
			float weight = NormalDistribution( standardDeviation );
			weightsHiddenOutput[weightIdx] = weight;
		}
	}
	public void Train()
	{
		// reset training state
		currentEpoch = 0;
		trainingSetAccuracy = 0;
		validationSetAccuracy = 0;
		generalizationSetAccuracy = 0;
		trainingSetMSE = 0;
		validationSetMSE = 0;
		generalizationSetMSE = 0;
		// load training data
		Console.Write( " Loading MNIST training data... " );
		trainingData = new TrainingData( 4000, 1000, 1000 );
		string path = "../data/";
		if (!File.Exists( path + "images.bin" )) path = "../../data/";
		using (BinaryReader b = new BinaryReader( File.Open( path + "images.bin", FileMode.Open ) ))
		{
			for( int i = 0; i < 4000; i++ ) 
			{
				trainingData.trainingSet.entry[i].inputs = new float[784];
				for( int j = 0; j < 784; j++ ) trainingData.trainingSet.entry[i].inputs[j] = b.ReadSingle();
			}
			for( int i = 0; i < 1000; i++ ) 
			{
				trainingData.generalizationSet.entry[i].inputs = new float[784];
				for( int j = 0; j < 784; j++ ) trainingData.generalizationSet.entry[i].inputs[j] = b.ReadSingle();
			}
			for( int i = 0; i < 1000; i++ ) 
			{
				trainingData.validationSet.entry[i].inputs = new float[784];
				for( int j = 0; j < 784; j++ ) trainingData.validationSet.entry[i].inputs[j] = b.ReadSingle();
			}
		}
		using (BinaryReader b = new BinaryReader( File.Open( path + "labels.bin", FileMode.Open ) ))
		{
			for( int i = 0; i < 4000; i++ ) 
			{
				trainingData.trainingSet.entry[i].expected = new int[10];
				for( int j = 0; j < 10; j++ ) trainingData.trainingSet.entry[i].expected[j] = b.ReadInt32();
			}
			for( int i = 0; i < 1000; i++ ) 
			{
				trainingData.generalizationSet.entry[i].expected = new int[10];
				for( int j = 0; j < 10; j++ ) trainingData.generalizationSet.entry[i].expected[j] = b.ReadInt32();
			}
			for( int i = 0; i < 1000; i++ ) 
			{
				trainingData.validationSet.entry[i].expected = new int[10];
				for( int j = 0; j < 10; j++ ) trainingData.validationSet.entry[i].expected[j] = b.ReadInt32();
			}
		}
		Console.Write( "done.\n" );
		// print header
		Console.Write( " Neural Network Training Starting: \n" );
		Console.Write( "==========================================================================\n" );
		Console.Write( " LR: " + LEARNINGRATE + " momentum: " + MOMENTUM + " max epochs: " + MAXEPOCHS + "\n" );
		Console.Write( " " + INPUTSIZE + " input neurons, " + NUMHIDDEN + " hidden neurons, " + NUMOUTPUT + " output neurons\n" );
		Console.Write( "==========================================================================\n" );
		// train network using training dataset for training and generalization dataset for testing
		while ((trainingSetAccuracy < TARGETACCURACY || generalizationSetAccuracy < TARGETACCURACY) && currentEpoch < MAXEPOCHS)
		{
			// use training set to train network
			timer.Restart();
			RunEpoch( ref trainingData.trainingSet );
			float elapsed = timer.ElapsedMilliseconds / 1000.0f;
			// get generalization set accuracy and MSE
			GetSetAccuracyAndMSE( ref trainingData.generalizationSet, ref generalizationSetAccuracy, ref generalizationSetMSE );
			Console.Write( "Epoch " + currentEpoch.ToString( "000" ) + " - TS acc " + trainingSetAccuracy.ToString( "00.00" ) );
			Console.Write( "%, MSE " + trainingSetMSE.ToString( "0.00000" ) + " GS acc " + generalizationSetAccuracy.ToString( "00.00" ) );
			Console.Write( "%, MSE " + generalizationSetMSE.ToString( "0.00000" ) + ", " + timer.ElapsedMilliseconds.ToString( "0.0" ) + "s\n" );
			currentEpoch++;
		}
		// get validation set accuracy and MSE
		GetSetAccuracyAndMSE( ref trainingData.validationSet, ref validationSetAccuracy, ref validationSetMSE );
		// print validation accuracy and MSE
		Console.Write( "\nTraining complete. Epochs: " + currentEpoch + "\n" );
		Console.Write( " Validation set accuracy: " + validationSetAccuracy + "\n Validation set MSE: " + validationSetMSE + "\n" );
	}
	void RunEpoch( ref TrainingSet set )
	{
		float incorrectEntries = 0;
		float MSE = 0;
		for( int i = 0; i < set.size; i++ )
		{
			// feed inputs through network and back propagate errors
			Evaluate( ref set.entry[i].inputs );
			Backpropagate( ref set.entry[i].expected );
			// check all outputs from neural network against desired values
			bool resultCorrect = true;
			for( int j = 0; j < NUMOUTPUT; j++ )
			{
				if (clampedOutputs[j] != set.entry[i].expected[j]) resultCorrect = false;
				float delta = outputNeurons[j] - set.entry[i].expected[j];
				MSE += delta * delta;
			}
			if (!resultCorrect) incorrectEntries++;
		}
		// update training accuracy and MSE
		trainingSetAccuracy = 100.0f - (incorrectEntries / set.size * 100.0f);
		trainingSetMSE = MSE / (NUMOUTPUT * set.size);
	}
	public int[] Evaluate( ref float[] input )
	{
		// set input values
		for( int i = 0; i < INPUTSIZE; i++ ) inputNeurons[i] = input[i];
		// update hidden neurons
		for( int i = 0; i < NUMHIDDEN; i++ )
		{
			hiddenNeurons[i] = 0;
			// get weighted sum of pattern and bias neuron
			for( int j = 0; j <= INPUTSIZE; j++ )
			{
				int weightIdx = GetInputHiddenWeightIndex( j, i );
				hiddenNeurons[i] += inputNeurons[j] * weightsInputHidden[weightIdx];
			}
			// apply activation function
			hiddenNeurons[i] = SigmoidActivationFunction( hiddenNeurons[i] );
		}
		// calculate output values - include bias neuron
		for( int i = 0; i < NUMOUTPUT; i++ )
		{
			outputNeurons[i] = 0;
			// get weighted sum of pattern and bias neuron
			for( int j = 0; j <= NUMHIDDEN; j++ )
			{
				int weightIdx = GetHiddenOutputWeightIndex( j, i );
				outputNeurons[i] += hiddenNeurons[j] * weightsHiddenOutput[weightIdx];
			}
			// apply activation function and clamp the result
			outputNeurons[i] = SigmoidActivationFunction( outputNeurons[i] );
			clampedOutputs[i] = ClampOutputValue( outputNeurons[i] );
		}
		return clampedOutputs;
	}
	void Backpropagate( ref int[] expectedOutputs )
	{
		// modify deltas between hidden and output layers
		for( int i = 0; i < NUMOUTPUT; i++ )
		{
			// get error gradient for every output node
			errorGradientsOutput[i] = GetOutputErrorGradient( (float)expectedOutputs[i], outputNeurons[i] );
			// for all nodes in hidden layer and bias neuron
			for( int j = 0; j <= NUMHIDDEN; j++ )
			{
				int weightIdx = GetHiddenOutputWeightIndex( j, i );
				// calculate change in weight
				deltaHiddenOutput[weightIdx] = LEARNINGRATE * hiddenNeurons[j] * errorGradientsOutput[i] + MOMENTUM * deltaHiddenOutput[weightIdx];
			}
		}
		// modify deltas between input and hidden layers
		for( int i = 0; i <= NUMHIDDEN; i++ )
		{
			// get error gradient for every hidden node
			errorGradientsHidden[i] = GetHiddenErrorGradient( i );
			// for all nodes in input layer and bias neuron
			for( int j = 0; j <= INPUTSIZE; j++ )
			{
				int weightIdx = GetInputHiddenWeightIndex( j, i );
				// calculate change in weight 
				deltaInputHidden[weightIdx] = LEARNINGRATE * inputNeurons[j] * errorGradientsHidden[i] + MOMENTUM * deltaInputHidden[weightIdx];
			}
		}
		// update the weights
		UpdateWeights();
	}
	float GetHiddenErrorGradient( int hiddenIdx )
	{
		// get sum of hidden->output weights * output error gradients
		float weightedSum = 0;
		for( int i = 0; i < NUMOUTPUT; i++ )
		{
			int weightIdx = GetHiddenOutputWeightIndex( hiddenIdx, i );
			weightedSum += weightsHiddenOutput[weightIdx] * errorGradientsOutput[i];
		}
		// return error gradient
		return hiddenNeurons[hiddenIdx] * (1.0f - hiddenNeurons[hiddenIdx]) * weightedSum;
	}
	void UpdateWeights()
	{
		// input -> hidden weights
		for( int i = 0; i <= INPUTSIZE; i++ ) for( int j = 0; j <= NUMHIDDEN; j++ )
		{
			int weightIdx = GetInputHiddenWeightIndex( i, j );
			weightsInputHidden[weightIdx] += deltaInputHidden[weightIdx];
		}
		// hidden -> output weights
		for( int i = 0; i <= NUMHIDDEN; i++ ) for ( int j = 0; j < NUMOUTPUT; j++ )
		{
			int weightIdx = GetHiddenOutputWeightIndex( i, j );
			weightsHiddenOutput[weightIdx] += deltaHiddenOutput[weightIdx];
		}
	}
	void GetSetAccuracyAndMSE( ref TrainingSet set, ref float accuracy, ref float MSE )
	{
		accuracy = 0.0f;
		MSE = 0.0f;
		float numIncorrectResults = 0;
		for( int i = 0; i < set.size; i++ )
		{
			Evaluate( ref set.entry[i].inputs );
			// check if the network outputs match the expected outputs
			int correctResults = 0;
			for( int j = 0; j < NUMOUTPUT; j++ )
			{
				correctResults += (clampedOutputs[j] == set.entry[i].expected[j]) ? 1 : 0;
				float delta = outputNeurons[j] - set.entry[i].expected[j];
				MSE += delta * delta;
			}
			if (correctResults != NUMOUTPUT) numIncorrectResults++;
		}
		accuracy = 100.0f - (numIncorrectResults / set.size * 100.0f);
		MSE = MSE / (NUMOUTPUT * set.size);
	}
}

} // namespace Template