﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;

namespace Template {

// data layout, neural net:
// 1. A bunch of consts
// 2. Three arrays storing the biases of the neurons:
//    - inputNeurons								785 floats
//    - hiddenNeurons								151 floats
//    - outputNeurons								10 floats
// 3. An array storing the clamped outputs			10 ints
// 4. Two arrays storing the links between these:
//    - weightsInputHidden							785 * 151 floats
//    - weightsHiddenOutput							151 * 10 floats
// training data layout:
// 1. TrainingData, 6000 entries:
//    - inputs										784 floats
//    - expected values								10 ints
// 2. deltas
//    - deltaInputHidden							785 * 151 floats
//    - deltaHiddenOutput							151 * 10 floats
// 3. gradients
//    - errorGradientsHidden						785 floats
//    - errorGradientsOutput						10 floats
// 4. training values
//    - currentEpoch;								1 int
//    - trainingSetAccuracy							1 float
//    - validationSetAccuracy						1 float
//    - generalizationSetAccuracy					1 float
//    - trainingSetMSE								1 float
//    - validationSetMSE							1 float
//    - generalizationSetMSE						1 float
// IDEA: store everything in three buffers, use constant offsets to
//       access specific data.
// A. The neural net:
//    inputNeurons				OFFS = 0
//    hiddenNeurons				OFFS = 800
//    outputNeurons				OFFS = 960 
//    clamped outputs			OFFS = 976 (int)
//    training values			OFFS = 992..998
//    weightsInputHidden		OFFS = 1024, 151 * 800 or 160 * 785
//    weightsHiddenOutput		OFFS = 124 * 1024, 160 * 10 or 151 * 16
//    (weights fit in 512Kb)
// B. TrainingData:
//    inputs					OFFS = idx * 800
//    expected					OFFS = idx * 800 + 784
//    (data fits in 19MB)
// C. deltas, gradients:
//    deltaInputHidden			OFFS = 0, 151 * 800 or 160 * 785
//    deltaHiddenOutput			OFFS = 124 * 1024, 160 * 10 or 151 * 16
//    errorGradientsHidden      OFFS = 127 * 1024
//    errorGradientsOutput		OFFS = 127 * 1024 + 800
//    (fits in 512Kb)
public struct TrainingEntry
{
	public float[] inputs;
	public int[] expected;
};
public struct TrainingSet
{
	public TrainingEntry[] entry;
	public int size;
};
public struct TrainingData
{
	public TrainingData( int t, int g, int v )
	{
		trainingSet.entry = new TrainingEntry[t];
		trainingSet.size = t;
		generalizationSet.entry = new TrainingEntry[g];
		generalizationSet.size = g;
		validationSet.entry = new TrainingEntry[v];
		validationSet.size = v;
	}
	public TrainingSet trainingSet;
	public TrainingSet generalizationSet;
	public TrainingSet validationSet;
};
public class NeuralNet
{
	// constants defining the topology of the net
	const int INPUTSIZE = 28 * 28;
	const int NUMHIDDEN = 150;
	const int NUMOUTPUT = 10;
	// constants steering training process
	const float LEARNINGRATE = 0.001f;
	const float MOMENTUM = 0.9f;
	const int MAXEPOCHS = 600;
	const float TARGETACCURACY = 95.0f;
	// pi
	const float PI = 3.14159265359f;
	// neural net storage
	/* float[] inputNeurons;
	float[] hiddenNeurons;
	float[] outputNeurons;
	int[] clampedOutputs;
	float[] weightsInputHidden;
	float[] weightsHiddenOutput;
	// training data
	TrainingData trainingData;
	float[] deltaInputHidden;		
	float[] deltaHiddenOutput;		
	float[] errorGradientsHidden;	
	float[] errorGradientsOutput;	
	int currentEpoch;			
	float trainingSetAccuracy;
	float validationSetAccuracy;
	float generalizationSetAccuracy;
	float trainingSetMSE;
	float validationSetMSE;
	float generalizationSetMSE; */

	// neural net data access
	int inputNeuron( int i ) { return i; }
	int hiddenNeuron( int i ) { return i + 800; }
	int outputNeuron( int i ) { return i + 960; }
	int	clampedOutput( int i ) { return i + 976; }
	int weightInputHidden( int i, int h ) { return 1024 + h * 800 + i; }
	int weightHiddenOutput( int h, int o ) { return 124 * 1024 + h * 10 + o; }
	int currentEpoch() { return 992; }
	int trainingSetAccuracy() { return 993; }
	int validationSetAccuracy() { return 994; }
	int generalizationSetAccuracy() { return 995; }
	int trainingSetMSE() { return 996; }
	int validationSetMSE() { return 997; }
	int generalizationSetMSE() { return 998; }

	// training data access
	int input( int i, int n ) { return i * 800 + n; }
	int expected( int i, int n ) { return i * 800 + 788 + n; }

	// delta / gradient data access
	int deltaInputHidden( int i, int h ) { return h * 800 + i; }
	int deltaHiddenOutput( int h, int o ) { return 118 * 1024 + h * 10 + o; }
	int errorGradientHidden( int i ) { return 127 * 1024 + i; }
	int errorGradientOutput( int i ) { return 127 * 1024 + 800 + i; }

	float[] ND = new float[128 * 1024];
	float[] TD = new float[4800000];
	float[] DG = new float[128 * 1024];
	// rng
	Random rd = new Random( 1000 );
	// timer
	Stopwatch timer = new Stopwatch();
	public NeuralNet()
	{
		InitializeNetwork();
		InitializeWeights();
	}
	int GetInputHiddenWeightIndex( int inputIdx, int hiddenIdx ) { return inputIdx * NUMHIDDEN + hiddenIdx; }
	int GetHiddenOutputWeightIndex( int hiddenIdx, int outputIdx ) { return hiddenIdx * NUMOUTPUT + outputIdx; }
	float SigmoidActivationFunction( float x ) { return 1.0f / (1.0f + (float)Math.Exp( -x )); }
	int ClampOutputValue( float x ) { if ( x < 0.1f ) return 0; else if ( x > 0.9f ) return 1; else return -1; }
	float GetOutputErrorGradient( float desiredValue, float outputValue ) { return outputValue * (1.0f - outputValue) * (desiredValue - outputValue); }
	void InitializeNetwork()
	{	
		/* const int totalNumInputs = INPUTSIZE + 1;
		const int totalNumHiddens = NUMHIDDEN + 1;
		inputNeurons = new float[totalNumInputs];
		hiddenNeurons = new float[totalNumHiddens];
		outputNeurons = new float[NUMOUTPUT];
		clampedOutputs = new int[NUMOUTPUT]; */
		// set bias values
		ND[inputNeuron( INPUTSIZE )] = -1.0f; // inputNeurons[INPUTSIZE] = -1.0f;
		ND[hiddenNeuron( NUMHIDDEN )] = -1.0f; // hiddenNeurons[NUMHIDDEN] = -1.0f;
		// allocate storage and initialize and layer weights
		/* weightsInputHidden = new float[totalNumInputs * totalNumHiddens];
		weightsHiddenOutput = new float[totalNumHiddens * NUMOUTPUT];
		// allocate storage for training
		deltaInputHidden = new float[totalNumInputs * totalNumHiddens];
		deltaHiddenOutput = new float[totalNumInputs * NUMOUTPUT];
		errorGradientsHidden = new float[totalNumHiddens];
		errorGradientsOutput = new float[NUMOUTPUT]; */
	}
	float NormalDistribution( double stddev )
	{
		double u1 = 1.0 - rd.NextDouble(), u2 = 1.0 - rd.NextDouble();
		double randStdNormal = Math.Sqrt( -2.0 * Math.Log( u1 ) ) * Math.Sin( 2.0 * Math.PI * u2 );
		return (float)(stddev * randStdNormal);
	}
	void InitializeWeights()
	{
		const float distributionRangeHalfWidth = 2.4f / INPUTSIZE;
		const float standardDeviation = distributionRangeHalfWidth * 2 / 6;
		// set weights to normally distributed random values between [-2.4 / numInputs, 2.4 / numInputs]
		for( int i = 0; i <= INPUTSIZE; i++ ) for( int j = 0; j < NUMHIDDEN; j++ )
		{
			// int weightIdx = GetInputHiddenWeightIndex( i, j );
			float weight = NormalDistribution( standardDeviation );
			ND[weightInputHidden( i, j )] = weight; // weightsInputHidden[weightIdx] = weight;
		}
		// set weights to normally distributed random values between [-2.4 / numInputs, 2.4 / numInputs]
		for( int i = 0; i <= NUMHIDDEN; i++ ) for( int j = 0; j < NUMOUTPUT; j++ )
		{
			// int weightIdx = GetHiddenOutputWeightIndex( i, j );
			float weight = NormalDistribution( standardDeviation );
			ND[weightHiddenOutput( i, j )] = weight; // weightsHiddenOutput[weightIdx] = weight;
		}
	}
	public void Train()
	{
		// reset training state
		ND[currentEpoch()] = 0; // currentEpoch = 0;
		ND[trainingSetAccuracy()] = 0; // trainingSetAccuracy = 0;
		ND[validationSetAccuracy()] = 0; // validationSetAccuracy = 0;
		ND[generalizationSetAccuracy()] = 0; // generalizationSetAccuracy = 0;
		ND[trainingSetMSE()] = 0; // trainingSetMSE = 0;
		ND[validationSetMSE()] = 0; // validationSetMSE = 0;
		ND[generalizationSetMSE()] = 0; // generalizationSetMSE = 0;
		// load training data
		Console.Write( " Loading MNIST training data... " );
		// trainingData = new TrainingData( 4000, 1000, 1000 );
		string path = "../data/";
		if (!File.Exists( path + "images.bin" )) path = "../../data/";
		using (BinaryReader b = new BinaryReader( File.Open( path + "images.bin", FileMode.Open ) ))
		{
			for( int i = 0; i < 6000; i++ ) 
			{
				for( int j = 0; j < 784; j++ ) TD[input( i, j )] = b.ReadSingle();
			}
			/* for( int i = 0; i < 4000; i++ ) 
			{
				trainingData.trainingSet.entry[i].inputs = new float[784];
				for( int j = 0; j < 784; j++ ) trainingData.trainingSet.entry[i].inputs[j] = b.ReadSingle();
			}
			for( int i = 0; i < 1000; i++ ) 
			{
				trainingData.generalizationSet.entry[i].inputs = new float[784];
				for( int j = 0; j < 784; j++ ) trainingData.generalizationSet.entry[i].inputs[j] = b.ReadSingle();
			}
			for( int i = 0; i < 1000; i++ ) 
			{
				trainingData.validationSet.entry[i].inputs = new float[784];
				for( int j = 0; j < 784; j++ ) trainingData.validationSet.entry[i].inputs[j] = b.ReadSingle();
			} */
		}
		using (BinaryReader b = new BinaryReader( File.Open( path + "labels.bin", FileMode.Open ) ))
		{
			for( int i = 0; i < 6000; i++ ) 
			{
				// trainingData.trainingSet.entry[i].expected = new int[10];
				for( int j = 0; j < 10; j++ ) TD[expected( i, j )] = b.ReadInt32();
			}
			/* for( int i = 0; i < 4000; i++ ) 
			{
				trainingData.trainingSet.entry[i].expected = new int[10];
				for( int j = 0; j < 10; j++ ) trainingData.trainingSet.entry[i].expected[j] = b.ReadInt32();
			}
			for( int i = 0; i < 1000; i++ ) 
			{
				trainingData.generalizationSet.entry[i].expected = new int[10];
				for( int j = 0; j < 10; j++ ) trainingData.generalizationSet.entry[i].expected[j] = b.ReadInt32();
			}
			for( int i = 0; i < 1000; i++ ) 
			{
				trainingData.validationSet.entry[i].expected = new int[10];
				for( int j = 0; j < 10; j++ ) trainingData.validationSet.entry[i].expected[j] = b.ReadInt32();
			} */
		}
		Console.Write( "done.\n" );
		// print header
		Console.Write( " Neural Network Training Starting: \n" );
		Console.Write( "==========================================================================\n" );
		Console.Write( " LR: " + LEARNINGRATE + " momentum: " + MOMENTUM + " max epochs: " + MAXEPOCHS + "\n" );
		Console.Write( " " + INPUTSIZE + " input neurons, " + NUMHIDDEN + " hidden neurons, " + NUMOUTPUT + " output neurons\n" );
		Console.Write( "==========================================================================\n" );
		// train network using training dataset for training and generalization dataset for testing
		while ((ND[trainingSetAccuracy()] < TARGETACCURACY || ND[generalizationSetAccuracy()] < TARGETACCURACY) && ND[currentEpoch()] < MAXEPOCHS)
		{
			// use training set to train network
			timer.Restart();
			RunEpoch( /* ref trainingData.trainingSet */ );
			float elapsed = timer.ElapsedMilliseconds / 1000.0f;
			// get generalization set accuracy and MSE
			GetSetAccuracyAndMSE( 4000, ref ND[generalizationSetAccuracy()], ref ND[generalizationSetMSE()] ); // GetSetAccuracyAndMSE( ref trainingData.generalizationSet, ref generalizationSetAccuracy, ref generalizationSetMSE );
			Console.Write( "Epoch " + ND[currentEpoch()].ToString( "000" ) + " - TS acc " + ND[trainingSetAccuracy()].ToString( "00.00" ) );
			Console.Write( "%, MSE " + ND[trainingSetMSE()].ToString( "0.00000" ) + " GS acc " + ND[generalizationSetAccuracy()].ToString( "00.00" ) );
			Console.Write( "%, MSE " + ND[generalizationSetMSE()].ToString( "0.00000" ) + ", " + timer.ElapsedMilliseconds.ToString( "0.0" ) + "s\n" );
			ND[currentEpoch()]++;
		}
		// get validation set accuracy and MSE
		GetSetAccuracyAndMSE( 5000, ref ND[validationSetAccuracy()], ref ND[validationSetMSE()] ); // GetSetAccuracyAndMSE( ref trainingData.validationSet, ref validationSetAccuracy, ref validationSetMSE );
		// print validation accuracy and MSE
		Console.Write( "\nTraining complete. Epochs: " + ND[currentEpoch()] + "\n" );
		Console.Write( " Validation set accuracy: " + ND[validationSetAccuracy()] + "\n Validation set MSE: " + ND[validationSetMSE()] + "\n" );
	}
	void RunEpoch( /* ref TrainingSet set */ )
	{
		float incorrectEntries = 0;
		float MSE = 0;
		for( int entryIdx = 0; entryIdx < 4000; entryIdx++ ) // for( int i = 0; i < set.size; i++ )
		{
			// feed inputs through network and back propagate errors
			Evaluate( entryIdx ); // Evaluate( ref set.entry[i].inputs );
			Backpropagate( entryIdx ); // Backpropagate( ref set.entry[i].expected );
			// check all outputs from neural network against desired values
			bool resultCorrect = true;
			for( int j = 0; j < NUMOUTPUT; j++ )
			{
				if (ND[clampedOutput( j )] != TD[expected( entryIdx, j )]) resultCorrect = false; // if (clampedOutputs[j] != set.entry[i].expected[j]) resultCorrect = false;
				float delta = ND[outputNeuron( j )] - (int)TD[expected( entryIdx, j )]; // float delta = outputNeurons[j] - set.entry[i].expected[j];
				MSE += delta * delta;
			}
			if (!resultCorrect) incorrectEntries++;
		}
		// update training accuracy and MSE
		ND[trainingSetAccuracy()] = 100.0f - (incorrectEntries / 4000 * 100.0f); // trainingSetAccuracy = 100.0f - (incorrectEntries / set.size * 100.0f);
		ND[trainingSetMSE()] = MSE / (NUMOUTPUT * 4000); // trainingSetMSE = MSE / (NUMOUTPUT * set.size);
	}
	public void /* int[] */ Evaluate( int entryIdx )
	{
		// set input values
		for( int i = 0; i < INPUTSIZE; i++ ) ND[inputNeuron( i )] = TD[input( entryIdx, i )]; // inputNeurons[i] = input[i];
		// update hidden neurons
		for( int i = 0; i < NUMHIDDEN; i++ )
		{
			ND[hiddenNeuron( i )] = 0; // hiddenNeurons[i] = 0;
			// get weighted sum of pattern and bias neuron
			for( int j = 0; j <= INPUTSIZE; j++ )
			{
				// int weightIdx = GetInputHiddenWeightIndex( j, i );
				ND[hiddenNeuron( i )] += ND[inputNeuron( j )] * ND[weightInputHidden( j, i )]; // hiddenNeurons[i] += inputNeurons[j] * weightsInputHidden[weightIdx];
			}
			// apply activation function
			ND[hiddenNeuron( i )] = SigmoidActivationFunction( ND[hiddenNeuron( i )] ); // hiddenNeurons[i] = SigmoidActivationFunction( hiddenNeurons[i] );
		}
		// calculate output values - include bias neuron
		for( int i = 0; i < NUMOUTPUT; i++ )
		{
			ND[outputNeuron( i )] = 0; // outputNeurons[i] = 0;
			// get weighted sum of pattern and bias neuron
			for( int j = 0; j <= NUMHIDDEN; j++ )
			{
				// int weightIdx = GetHiddenOutputWeightIndex( j, i );
				ND[outputNeuron( i )] += ND[hiddenNeuron( j )] * ND[weightHiddenOutput( j, i )]; // outputNeurons[i] += hiddenNeurons[j] * weightsHiddenOutput[weightIdx];
			}
			// apply activation function and clamp the result
			ND[outputNeuron( i )] = SigmoidActivationFunction( ND[outputNeuron( i )] ); // outputNeurons[i] = SigmoidActivationFunction( outputNeurons[i] );
			ND[clampedOutput( i )] = ClampOutputValue( ND[outputNeuron( i )] ); // clampedOutputs[i] = ClampOutputValue( outputNeurons[i] );
		}
		// return clampedOutputs;
	}
	void Backpropagate( int entryIdx /* ref int[] expectedOutputs */ )
	{
		// modify deltas between hidden and output layers
		for( int i = 0; i < NUMOUTPUT; i++ )
		{
			// get error gradient for every output node
			DG[errorGradientOutput( i )] = GetOutputErrorGradient( TD[expected( entryIdx, i )], ND[outputNeuron( i )] ); // errorGradientsOutput[i] = GetOutputErrorGradient( (float)expectedOutputs[i], outputNeurons[i] );
			// for all nodes in hidden layer and bias neuron
			for( int j = 0; j <= NUMHIDDEN; j++ )
			{
				// int weightIdx = GetHiddenOutputWeightIndex( j, i );
				// calculate change in weight
				DG[deltaHiddenOutput( j, i )] = LEARNINGRATE * ND[hiddenNeuron( j )] * DG[errorGradientOutput( i )] + MOMENTUM * DG[deltaHiddenOutput( j, i )]; // deltaHiddenOutput[weightIdx] = LEARNINGRATE * hiddenNeurons[j] * errorGradientsOutput[i] + MOMENTUM * deltaHiddenOutput[weightIdx];
			}
		}
		// modify deltas between input and hidden layers
		for( int i = 0; i <= NUMHIDDEN; i++ )
		{
			// get error gradient for every hidden node
			DG[errorGradientHidden( i )] = GetHiddenErrorGradient( i ); // errorGradientsHidden[i] = GetHiddenErrorGradient( i );
			// for all nodes in input layer and bias neuron
			for( int j = 0; j <= INPUTSIZE; j++ )
			{
				// int weightIdx = GetInputHiddenWeightIndex( j, i );
				// calculate change in weight 
				DG[deltaInputHidden( j, i )] = LEARNINGRATE * ND[inputNeuron( j )] * DG[errorGradientHidden( i )] + MOMENTUM * DG[deltaInputHidden( j, i )]; // deltaInputHidden[weightIdx] = LEARNINGRATE * inputNeurons[j] * errorGradientsHidden[i] + MOMENTUM * deltaInputHidden[weightIdx];
			}
		}
		// update the weights
		UpdateWeights();
	}
	float GetHiddenErrorGradient( int hiddenIdx )
	{
		// get sum of hidden->output weights * output error gradients
		float weightedSum = 0;
		for( int i = 0; i < NUMOUTPUT; i++ )
		{
			// int weightIdx = GetHiddenOutputWeightIndex( hiddenIdx, i );
			weightedSum += ND[weightHiddenOutput( hiddenIdx, i )] * DG[errorGradientOutput( i )]; // weightedSum += weightsHiddenOutput[weightIdx] * errorGradientsOutput[i];
		}
		// return error gradient
		return ND[hiddenNeuron( hiddenIdx )] * (1.0f - ND[hiddenNeuron( hiddenIdx )]) * weightedSum; // return hiddenNeurons[hiddenIdx] * (1.0f - hiddenNeurons[hiddenIdx]) * weightedSum;
	}
	void UpdateWeights()
	{
		// input -> hidden weights
		for( int i = 0; i <= INPUTSIZE; i++ ) for( int j = 0; j <= NUMHIDDEN; j++ )
		{
			// int weightIdx = GetInputHiddenWeightIndex( i, j );
			ND[weightInputHidden( i, j )] += DG[deltaInputHidden( i, j )]; // weightsInputHidden[weightIdx] += deltaInputHidden[weightIdx];
		}
		// hidden -> output weights
		for( int i = 0; i <= NUMHIDDEN; i++ ) for ( int j = 0; j < NUMOUTPUT; j++ )
		{
			// int weightIdx = GetHiddenOutputWeightIndex( i, j );
			ND[weightHiddenOutput( i, j )] += DG[deltaHiddenOutput( i, j )]; // weightsHiddenOutput[weightIdx] += deltaHiddenOutput[weightIdx];
		}
	}
	void GetSetAccuracyAndMSE( int firstIdx, /* ref TrainingSet set, */ ref float accuracy, ref float MSE )
	{
		accuracy = 0.0f;
		MSE = 0.0f;
		float numIncorrectResults = 0;
		for( int i = 0; i < 1000; i++ ) // for( int i = 0; i < set.size; i++ )
		{
			Evaluate( firstIdx + i );
			// check if the network outputs match the expected outputs
			int correctResults = 0;
			for( int j = 0; j < NUMOUTPUT; j++ )
			{
				correctResults += (ND[clampedOutput( j )] == TD[expected( firstIdx + i, j )]) ? 1 : 0; // correctResults += (clampedOutputs[j] == set.entry[i].expected[j]) ? 1 : 0;
				float delta = ND[outputNeuron( j )] - (int)TD[expected( firstIdx + i, j )]; // float delta = outputNeurons[j] - set.entry[i].expected[j];
				MSE += delta * delta;
			}
			if (correctResults != NUMOUTPUT) numIncorrectResults++;
		}
		accuracy = 100.0f - (numIncorrectResults / 1000 * 100.0f);
		MSE = MSE / (NUMOUTPUT * 1000);
	}
}

} // namespace Template
