﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.IO;
using System.Threading;
using System.Threading.Tasks;

namespace Template {

public class NeuralNet
{
	// constants defining the topology of the net
	const int INPUTSIZE = 28 * 28;
	const int NUMHIDDEN = 150;
	const int NUMOUTPUT = 10;
	// constants steering training process
	const float LEARNINGRATE = 0.001f;
	const float MOMENTUM = 0.9f;
	const int MAXEPOCHS = 600;
	const float TARGETACCURACY = 95.0f;
	// pi
	const float PI = 3.14159265359f;
	// neural net storage
	float[] ND = new float[128 * 1024];
	float[] TD = new float[4800000];
	float[] DG = new float[128 * 1024];	
	// neural net data access
	int inputNeuron( int i ) { return i; }
	int hiddenNeuron( int i ) { return i + 800; }
	int outputNeuron( int i ) { return i + 960; }
	int	clampedOutput( int i ) { return i + 976; }
	int weightInputHidden( int i, int h ) { return 1024 + h * 800 + i; }
	int weightHiddenOutput( int h, int o ) { return 124 * 1024 + h * 10 + o; }
	int currentEpoch() { return 992; }
	int trainingSetAccuracy() { return 993; }
	int validationSetAccuracy() { return 994; }
	int generalizationSetAccuracy() { return 995; }
	int trainingSetMSE() { return 996; }
	int validationSetMSE() { return 997; }
	int generalizationSetMSE() { return 998; }
	// training data access
	int input( int i, int n ) { return i * 800 + n; }
	int expected( int i, int n ) { return i * 800 + 788 + n; }
	// delta / gradient data access
	int deltaInputHidden( int i, int h ) { return h * 800 + i; }
	int deltaHiddenOutput( int h, int o ) { return 118 * 1024 + h * 10 + o; }
	int	errorGradientHidden( int i ) { return 127 * 1024 + i; }
	int errorGradientOutput( int i ) { return 127 * 1024 + 800 + i; }
	// rng
	Random rd = new Random( 1000 );
	// timer
	Stopwatch timer = new Stopwatch();
	public NeuralNet()
	{
		InitializeNetwork();
		InitializeWeights();
	}
	int GetInputHiddenWeightIndex( int inputIdx, int hiddenIdx ) { return inputIdx * NUMHIDDEN + hiddenIdx; }
	int GetHiddenOutputWeightIndex( int hiddenIdx, int outputIdx ) { return hiddenIdx * NUMOUTPUT + outputIdx; }
	float SigmoidActivationFunction( float x ) { return 1.0f / (1.0f + (float)Math.Exp( -x )); }
	int ClampOutputValue( float x ) { if ( x < 0.1f ) return 0; else if ( x > 0.9f ) return 1; else return -1; }
	float GetOutputErrorGradient( float desiredValue, float outputValue ) { return outputValue * (1.0f - outputValue) * (desiredValue - outputValue); }
	void InitializeNetwork()
	{	
		// set bias values
		ND[inputNeuron( INPUTSIZE )] = -1.0f;
		ND[hiddenNeuron( NUMHIDDEN )] = -1.0f;
	}
	float NormalDistribution( double stddev )
	{
		double u1 = 1.0 - rd.NextDouble(), u2 = 1.0 - rd.NextDouble();
		double randStdNormal = Math.Sqrt( -2.0 * Math.Log( u1 ) ) * Math.Sin( 2.0 * Math.PI * u2 );
		return (float)(stddev * randStdNormal);
	}
	void InitializeWeights()
	{
		const float distributionRangeHalfWidth = 2.4f / INPUTSIZE;
		const float standardDeviation = distributionRangeHalfWidth * 2 / 6;
		// set weights to normally distributed random values between [-2.4 / numInputs, 2.4 / numInputs]
		for( int i = 0; i <= INPUTSIZE; i++ ) for( int j = 0; j < NUMHIDDEN; j++ )
		{
			float weight = NormalDistribution( standardDeviation );
			ND[weightInputHidden( i, j )] = weight;
		}
		// set weights to normally distributed random values between [-2.4 / numInputs, 2.4 / numInputs]
		for( int i = 0; i <= NUMHIDDEN; i++ ) for( int j = 0; j < NUMOUTPUT; j++ )
		{
			float weight = NormalDistribution( standardDeviation );
			ND[weightHiddenOutput( i, j )] = weight;
		}
	}
	public void Train()
	{
		// reset training state
		ND[currentEpoch()] = 0;
		ND[trainingSetAccuracy()] = 0;
		ND[validationSetAccuracy()] = 0;
		ND[generalizationSetAccuracy()] = 0;
		ND[trainingSetMSE()] = 0;
		ND[validationSetMSE()] = 0;
		ND[generalizationSetMSE()] = 0;
		// load training data
		Console.Write( " Loading MNIST training data... " );
		string path = "../data/";
		if (!File.Exists( path + "images.bin" )) path = "../../data/";
		using (BinaryReader b = new BinaryReader( File.Open( path + "images.bin", FileMode.Open ) ))
		{
			for( int i = 0; i < 6000; i++ ) for( int j = 0; j < 784; j++ ) TD[input( i, j )] = b.ReadSingle();
		}
		using (BinaryReader b = new BinaryReader( File.Open( path + "labels.bin", FileMode.Open ) ))
		{
			for( int i = 0; i < 6000; i++ ) for( int j = 0; j < 10; j++ ) TD[expected( i, j )] = b.ReadInt32();
		}
		Console.Write( "done.\n" );
		// print header
		Console.Write( " Neural Network Training Starting: \n" );
		Console.Write( "==========================================================================\n" );
		Console.Write( " LR: " + LEARNINGRATE + " momentum: " + MOMENTUM + " max epochs: " + MAXEPOCHS + "\n" );
		Console.Write( " " + INPUTSIZE + " input neurons, " + NUMHIDDEN + " hidden neurons, " + NUMOUTPUT + " output neurons\n" );
		Console.Write( "==========================================================================\n" );
		// train network using training dataset for training and generalization dataset for testing
		while ((ND[trainingSetAccuracy()] < TARGETACCURACY || ND[generalizationSetAccuracy()] < TARGETACCURACY) && ND[currentEpoch()] < MAXEPOCHS)
		{
			// use training set to train network
			timer.Restart();
			RunEpoch();
			float elapsed = timer.ElapsedMilliseconds / 1000.0f;
			// get generalization set accuracy and MSE
			GetSetAccuracyAndMSE( 4000, ref ND[generalizationSetAccuracy()], ref ND[generalizationSetMSE()] );
			Console.Write( "Epoch " + ND[currentEpoch()].ToString( "000" ) + " - TS acc " + ND[trainingSetAccuracy()].ToString( "00.00" ) );
			Console.Write( "%, MSE " + ND[trainingSetMSE()].ToString( "0.00000" ) + " GS acc " + ND[generalizationSetAccuracy()].ToString( "00.00" ) );
			Console.Write( "%, MSE " + ND[generalizationSetMSE()].ToString( "0.00000" ) + ", " + timer.ElapsedMilliseconds.ToString( "0.0" ) + "s\n" );
			ND[currentEpoch()]++;
		}
		// get validation set accuracy and MSE
		GetSetAccuracyAndMSE( 5000, ref ND[validationSetAccuracy()], ref ND[validationSetMSE()] );
		// print validation accuracy and MSE
		Console.Write( "\nTraining complete. Epochs: " + ND[currentEpoch()] + "\n" );
		Console.Write( " Validation set accuracy: " + ND[validationSetAccuracy()] + "\n Validation set MSE: " + ND[validationSetMSE()] + "\n" );
	}
	void RunEpoch()
	{
		float incorrectEntries = 0;
		float MSE = 0;
		for( int entryIdx = 0; entryIdx < 4000; entryIdx++ )
		{
			// feed inputs through network and back propagate errors
			Evaluate( entryIdx );
			Backpropagate( entryIdx );
			// check all outputs from neural network against desired values
			bool resultCorrect = true;
			for( int j = 0; j < NUMOUTPUT; j++ )
			{
				if (ND[clampedOutput( j )] != TD[expected( entryIdx, j )]) resultCorrect = false;
				float delta = ND[outputNeuron( j )] - (int)TD[expected( entryIdx, j )];
				MSE += delta * delta;
			}
			if (!resultCorrect) incorrectEntries++;
		}
		// update training accuracy and MSE
		ND[trainingSetAccuracy()] = 100.0f - (incorrectEntries / 4000 * 100.0f);
		ND[trainingSetMSE()] = MSE / (NUMOUTPUT * 4000);
	}
	public void Evaluate( int entryIdx )
	{
		// set input values
		for( int i = 0; i < INPUTSIZE; i++ ) ND[inputNeuron( i )] = TD[input( entryIdx, i )];
		// update hidden neurons
		for( int i = 0; i < NUMHIDDEN; i++ ) ND[hiddenNeuron( i )] = 0;
		Parallel.For( 0, NUMHIDDEN, i =>
		{ 
			// get weighted sum of pattern and bias neuron
			for( int j = 0; j <= INPUTSIZE; j++ ) ND[hiddenNeuron( i )] += ND[inputNeuron( j )] * ND[weightInputHidden( j, i )];
			// apply activation function
			ND[hiddenNeuron( i )] = SigmoidActivationFunction( ND[hiddenNeuron( i )] );
		} );
		// calculate output values - include bias neuron
		for( int i = 0; i < NUMOUTPUT; i++ ) ND[outputNeuron( i )] = 0;
		for( int j = 0; j <= NUMHIDDEN; j++ ) for( int i = 0; i < NUMOUTPUT; i++ ) // can't (easily) do in parallel; fights for outputNeuron[i]
		{
			// get weighted sum of pattern and bias neuron
			ND[outputNeuron( i )] += ND[hiddenNeuron( j )] * ND[weightHiddenOutput( j, i )];
		}
		for( int i = 0; i < NUMOUTPUT; i++ )
		{
			// apply activation function and clamp the result
			ND[outputNeuron( i )] = SigmoidActivationFunction( ND[outputNeuron( i )] );
			ND[clampedOutput( i )] = ClampOutputValue( ND[outputNeuron( i )] );
		}
	}
	void Backpropagate( int entryIdx )
	{
		// modify deltas between hidden and output layers
		for( int i = 0; i < NUMOUTPUT; i++ )
		{
			// get error gradient for every output node
			DG[errorGradientOutput( i )] = GetOutputErrorGradient( TD[expected( entryIdx, i )], ND[outputNeuron( i )] );
		}
		// modify deltas between input and hidden layers
		Parallel.For( 0, NUMHIDDEN + 1, i => 
		{
			for( int k = 0; k < NUMOUTPUT; k++ )
			{
				// calculate change in weight
				DG[deltaHiddenOutput( i, k )] = LEARNINGRATE * ND[hiddenNeuron( i )] * DG[errorGradientOutput( k )] + MOMENTUM * DG[deltaHiddenOutput( i, k )];
			}
			// get error gradient for every hidden node
			DG[errorGradientHidden( i )] = GetHiddenErrorGradient( i );
			// for all nodes in input layer and bias neuron
			for( int j = 0; j <= INPUTSIZE; j++ )
			{
				// calculate change in weight 
				DG[deltaInputHidden( j, i )] = LEARNINGRATE * ND[inputNeuron( j )] * DG[errorGradientHidden( i )] + MOMENTUM * DG[deltaInputHidden( j, i )];
			}
		} );
		// update the weights
		UpdateWeights();
	}
	float GetHiddenErrorGradient( int hiddenIdx )
	{
		// get sum of hidden->output weights * output error gradients
		float weightedSum = 0;
		for( int i = 0; i < NUMOUTPUT; i++ )
			weightedSum += ND[weightHiddenOutput( hiddenIdx, i )] * DG[errorGradientOutput( i )];
		// return error gradient
		return ND[hiddenNeuron( hiddenIdx )] * (1.0f - ND[hiddenNeuron( hiddenIdx )]) * weightedSum;
	}
	void UpdateWeights()
	{
		// input -> hidden weights
		Parallel.For( 0, INPUTSIZE + 1, i =>
		{ 
			for( int j = 0; j <= NUMHIDDEN; j++ )
				ND[weightInputHidden( i, j )] += DG[deltaInputHidden( i, j )];
		} );
		// hidden -> output weights
		Parallel.For( 0, NUMHIDDEN + 1, i =>
		{
			for ( int j = 0; j < NUMOUTPUT; j++ )
				ND[weightHiddenOutput( i, j )] += DG[deltaHiddenOutput( i, j )];
		} );
	}
	void GetSetAccuracyAndMSE( int firstIdx, ref float accuracy, ref float MSE )
	{
		accuracy = 0.0f;
		MSE = 0.0f;
		float numIncorrectResults = 0;
		for( int i = 0; i < 1000; i++ )
		{
			Evaluate( firstIdx + i );
			// check if the network outputs match the expected outputs
			int correctResults = 0;
			for( int j = 0; j < NUMOUTPUT; j++ )
			{
				correctResults += (ND[clampedOutput( j )] == TD[expected( firstIdx + i, j )]) ? 1 : 0;
				float delta = ND[outputNeuron( j )] - (int)TD[expected( firstIdx + i, j )];
				MSE += delta * delta;
			}
			if (correctResults != NUMOUTPUT) numIncorrectResults++;
		}
		accuracy = 100.0f - (numIncorrectResults / 1000 * 100.0f);
		MSE = MSE / (NUMOUTPUT * 1000);
	}
}

} // namespace Template