// constants
#define INPUTSIZE		(28*28)
#define NUMHIDDEN		150
#define NUMOUTPUT		10
#define LEARNINGRATE	0.001f
#define MOMENTUM		0.9f

// neural net data access
int inputNeuron( int i ) { return i; }
int hiddenNeuron( int i ) { return i + 800; }
int outputNeuron( int i ) { return i + 960; }
int	clampedOutput( int i ) { return i + 976; }
int weightInputHidden( int i, int h ) { return 1024 + h * 800 + i; }
int weightHiddenOutput( int h, int o ) { return 124 * 1024 + h * 10 + o; }
int currentEpoch() { return 992; }
int trainingSetAccuracy() { return 993; }
int validationSetAccuracy() { return 994; }
int generalizationSetAccuracy() { return 995; }
int trainingSetMSE() { return 996; }
int validationSetMSE() { return 997; }
int generalizationSetMSE() { return 998; }

// training data access
int input( int i, int n ) { return i * 800 + n; }
int expected( int i, int n ) { return i * 800 + 788 + n; }

// delta / gradient data access
int deltaInputHidden( int i, int h ) { return h * 800 + i; }
int deltaHiddenOutput( int h, int o ) { return 118 * 1024 + h * 10 + o; }
int	errorGradientHidden( int i ) { return 127 * 1024 + i; }
int errorGradientOutput( int i ) { return 127 * 1024 + 800 + i; }

// helper functions
float SigmoidActivationFunction( float x ) { return 1.0f / (1.0f + exp( -x )); }
int ClampOutputValue( float x ) { if ( x < 0.1f ) return 0; else if ( x > 0.9f ) return 1; else return -1; }
float GetOutputErrorGradient( float desiredValue, float outputValue ) { return outputValue * (1.0f - outputValue) * (desiredValue - outputValue); }

// atomic add for floats
inline void AtomicAddFloat( volatile __global float* source, const float operand ) 
{
	// Hamuraru and Hindriksen, streamhpc.com
	union { unsigned int intVal; float floatVal; } newVal;
	union { unsigned int intVal; float floatVal; } prevVal;
	do 
	{
		prevVal.floatVal = *source;
		newVal.floatVal = prevVal.floatVal + operand;
	} 
	while (atomic_cmpxchg( (volatile __global unsigned int*)source, prevVal.intVal, newVal.intVal ) != prevVal.intVal);
}

// kernels

__kernel void Evaluate1( const int entryIdx, __global float* ND, __global float* TD )
{
	int id = get_global_id( 0 ); // get thread id, will be 0..INPUTSIZE-1
	// set input values
	/* for( int i = 0; i < INPUTSIZE; i++ ) */ ND[inputNeuron( id )] = TD[input( entryIdx, id )];
	// update hidden neurons
	/* for( int i = 0; i < NUMHIDDEN; i++ ) */ if (id < NUMHIDDEN) ND[hiddenNeuron( id )] = 0;
	/* for( int i = 0; i < NUMOUTPUT; i++ ) */ if (id < NUMOUTPUT) ND[outputNeuron( id )] = 0;
	// above code must complete before we go into Evaluate2
}

__kernel void Evaluate2( __global float* ND )
{
	int id = get_global_id( 0 ); // get thread id, will be 0..NUMHIDDEN-1
	/* Parallel.For( 0, NUMHIDDEN, i => { */
	// get weighted sum of pattern and bias neuron
	for( int j = 0; j <= INPUTSIZE; j++ ) ND[hiddenNeuron( id )] += ND[inputNeuron( j )] * ND[weightInputHidden( j, id )];
	// apply activation function
	ND[hiddenNeuron( id )] = SigmoidActivationFunction( ND[hiddenNeuron( id )] );
	// above code must complete before we go into Evaluate3
}

__kernel void Evaluate3( __global float* ND )
{
	int id = get_global_id( 0 ); // get thread id, will be 0..NUMHIDDEN
	// calculate output values - include bias neuron
	/* for( int j = 0; j <= NUMHIDDEN; j++ ) */ for( int i = 0; i < NUMOUTPUT; i++ )
	{
		// get weighted sum of pattern and bias neuron
		AtomicAddFloat( &ND[outputNeuron( i )], ND[hiddenNeuron( id )] * ND[weightHiddenOutput( id, i )] );
		// ND[outputNeuron( i )] += ND[hiddenNeuron( id )] * ND[weightHiddenOutput( id, i )];
	}
	// above code must complete before we go into Evaluate4
}

__kernel void Evaluate4( __global float* ND )
{
	int id = get_global_id( 0 ); // get thread id, will be 0..NUMOUTPUT
	/* for( int i = 0; i < NUMOUTPUT; i++ ) */
	{
		// apply activation function and clamp the result
		ND[outputNeuron( id )] = SigmoidActivationFunction( ND[outputNeuron( id )] );
		ND[clampedOutput( id )] = ClampOutputValue( ND[outputNeuron( id )] );
	}
}

float GetHiddenErrorGradient( int hiddenIdx, __global float* ND, __global float* DG )
{
	// get sum of hidden->output weights * output error gradients
	float weightedSum = 0;
	for( int i = 0; i < NUMOUTPUT; i++ ) weightedSum += ND[weightHiddenOutput( hiddenIdx, i )] * DG[errorGradientOutput( i )];
	// return error gradient
	return ND[hiddenNeuron( hiddenIdx )] * (1.0f - ND[hiddenNeuron( hiddenIdx )]) * weightedSum;
}

__kernel void Backpropagate1( int entryIdx, __global float* ND, __global float* TD, __global float* DG )
{
	int id = get_global_id( 0 ); // get thread id, will be 0..NUMOUTPUT
	// modify deltas between hidden and output layers
	/* for( int i = 0; i < NUMOUTPUT; i++ ) */
	{
		// get error gradient for every output node
		DG[errorGradientOutput( id )] = GetOutputErrorGradient( TD[expected( entryIdx, id )], ND[outputNeuron( id )] );
	}
}

__kernel void Backpropagate2( __global float* ND, __global float* DG )
{
	int id = get_global_id( 0 ); // get thread id, will be 0..NUMHIDDEN
	// modify deltas between input and hidden layers
	/* Parallel.For( 0, NUMHIDDEN + 1, i => { */
	for( int k = 0; k < NUMOUTPUT; k++ )
	{
		// calculate change in weight
		DG[deltaHiddenOutput( id, k )] = LEARNINGRATE * ND[hiddenNeuron( id )] * DG[errorGradientOutput( k )] + MOMENTUM * DG[deltaHiddenOutput( id, k )];
	}
	// get error gradient for every hidden node
	DG[errorGradientHidden( id )] = GetHiddenErrorGradient( id, ND, DG );
	// for all nodes in input layer and bias neuron
	for( int j = 0; j <= INPUTSIZE; j++ )
	{
		// calculate change in weight 
		DG[deltaInputHidden( j, id )] = LEARNINGRATE * ND[inputNeuron( j )] * DG[errorGradientHidden( id )] + MOMENTUM * DG[deltaInputHidden( j, id )];
	}
}

__kernel void UpdateWeights( __global float* ND, __global float* DG )
{
	int id = get_global_id( 0 ); // get thread id, will be 0..INPUTSIZE
	// input -> hidden weights
	/* Parallel.For( 0, INPUTSIZE + 1, i => { */
	for( int j = 0; j <= NUMHIDDEN; j++ ) ND[weightInputHidden( id, j )] += DG[deltaInputHidden( id, j )];
	// hidden -> output weights
	/* Parallel.For( 0, NUMHIDDEN + 1, i => { */
	if (id <= NUMHIDDEN) for ( int j = 0; j < NUMOUTPUT; j++ ) ND[weightHiddenOutput( id, j )] += DG[deltaHiddenOutput( id, j )];
}